"""PyTorch Matrix Product Operators for Neural Network Compression."""

from torch_mpo.layers import TTConv2d, TTLinear
from torch_mpo.utils import compress_model

__version__ = "0.1.0"
__all__ = ["TTLinear", "TTConv2d", "compress_model"]
